#!/usr/bin/python
"""idifference.

Generates a report about what's different between two disk images.

Process:

1. A map is kept of all filenames->sha1 codes and all sha1->filenames.
2. For each image, read all of the fileobject objects:
    - create new maps
    - Note when things change.
    - Delete each file in the old map as it is processed.
3. Report files left in map; that's the files that were deleted!
4. Replace the old maps with the new maps
"""

import sys
import fiwalk

# Allow nmaed pipes, character devices, block devices, regular files,
# links, symbolic links, sockets, shadow inodes (solaris), and whiteout
relevant_file_types = set(['p','c','b','r','l','s','h','w'])

def ptime(t):
    import time
    lt = time.localtime(t)
    return "%04d-%02d-%02d %02d:%02d:%02d" % (lt.tm_year,lt.tm_mon,lt.tm_mday,lt.tm_hour,lt.tm_min,lt.tm_sec)
    

def header():
    if options.html:
        print '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"',\
            '"http://www.w3.org/TR/html4/loose.dtd">'
        print '<html>'
        print '<body>'
        print '<style>'
        print 'body  { font-family: Sans-serif;}'
        print '.sha1 { font-family: monospace; font-size: small;}'
        print '.filesize { padding-left: 15px; padding-right: 15px; text-align: right;}'
        print '</style>'

def h1(title):
    global options
    if options.html:
        print "<h1>%s</h1>" % title
        return
    print "\n\n",title,"\n"

def h2(title):
    global options
    if options.html:
        print "<h2>%s</h2>" % title
        return
    print "\n",title,"\n"


def table(rows,styles=None):
    import sys
    global options
    def alldigits(x):
        if type(x)!=str and type(x)!=unicode: return False
        for ch in x:
            if ch.isdigit()==False: return False
        return True

    def fmt(x):
        if x==None: return ""
        if type(x)==int: return "%12d" % x
        if alldigits(x): return "%12d" % int(x)
        if type(x)==unicode: return x
        return unicode(x)
            
    if options.html:
        print "<table>"
        for row in rows:
            print "<tr>"
            if not styles:
                styles = [""]*len(rows)
            for (col,style) in zip(row,styles):
                print "<td class='%s'>%s</td>" % (style,col),
            print "<tr>"
        print "</table>"
        return
    for row in rows:
        try:
            line = u"\t".join([fmt(col) for col in row])
            print line
        except UnicodeEncodeError:
            # Fall back to manual join
            for col in row:
                for ch in fmt(col):
                    try: 
                        sys.stdout.write(ch)
                    except UnicodeEncodeError:
                        sys.stdout.write('?');
                sys.stdout.write("\t")
            print "(UNICODE ERROR)"

#
# This program keeps track of the current and previous diskstate in a single
# object called "DiskState". Another way to do that would have been to have
# the instance built from the XML file and then have another function that compares
# them.
#        

class DiskState:
    def __init__(self):
        self.new_sha1s = dict()
        self.new_fnames = dict()
        self.new_inodes = dict()
        self.next()
        
    def next(self):
        """Called when the next image is processed."""
        self.sha1s  = self.new_sha1s
        self.fnames = self.new_fnames
        self.inodes = self.new_inodes
        self.new_sha1s  = dict()
        self.new_fnames = dict()
        self.new_inodes = dict()
        self.new_files      = []
        self.renamed_files  = []
        self.changed_content  = []
        self.changed_properties = []
        self.timeline = []

    def process_fi(self,fi):
        # See if the filename changed its hash code
        changed = False
        if not fi.allocated(): return # only look at allocated files
        if fi.name_type() not in relevant_file_types: return        # only relevant file types
        if fi.filename()=='$FAT1' or fi.filename()=='$FAT2': return # ignore this
        if fi.name_type()=='-':
            print "unknown fi type: ",fi

        # Remember the file for the next generation
        self.new_sha1s[fi.sha1()] = fi
        self.new_fnames[fi.filename()] = fi
        self.new_inodes[fi.partition() + "/" + fi.inode()] = fi

        # See if a file changed or was resized
        ofi = self.fnames.get(fi.filename(),None)
        if ofi:
            if ofi.sha1()!=fi.sha1():
                self.changed_content.append((ofi,fi))
            elif ofi.atime() != fi.atime() or \
                    ofi.mtime() != fi.mtime() or \
                    ofi.crtime() != fi.crtime() or \
                    ofi.ctime() != fi.ctime():
                self.changed_properties.append((ofi,fi))

        # If a new file, note that (and optionally add to the timeline)
        if not ofi:
            self.new_files.append(fi)
            if options.timeline:
                create_time = fi.crtime()
                if not create_time: create_time = fi.ctime()
                if not create_time: create_time = fi.mtime()
                self.timeline.append((create_time,fi.filename(),"created"))

        # Delete files we have seen (so we can find out the files that were deleted)
        if fi.filename() in self.fnames:
            del self.fnames[fi.filename()]

        # Look for files that were renamed
        ofi = self.inodes.get(fi.partition() + "/" + fi.inode(),None)
        if ofi and ofi.filename() != fi.filename() and ofi.sha1()==fi.sha1():
            self.renamed_files.append((ofi,fi))


    def process(self,fname):
        self.current_fname = fname
        if fname.endswith(".xml"):
            fiwalk.fiwalk_using_sax(xmlfile=open(infile), flags=fiwalk.ALLOC_ONLY, callback=self.process_fi)
        else:
            fiwalk.fiwalk_using_sax(imagefile=open(infile), flags=fiwalk.ALLOC_ONLY, callback=self.process_fi)

    def print_fis(self,title,fis):
        h2(title)
        def fidate(mtime):
            try:
                return ptime(fi.mtime())
            except TypeError:
                return "n/a"
        res = [(fidate(fi),str(fi.filesize()),fi.filename()) for fi in fis]
        if res:
            res.sort()
            table(res)

    def print_fi2(self,title,fi2s):
        def prtime(t):
            return "%d (%s)" % (t,ptime(t))

        h2(title)
        res = []
        for(ofi,fi) in fi2s:
            if ofi.filename() != fi.filename():
                res.append((ofi.filename(),"renamed to",fi.filename()))
            if ofi.filesize() != fi.filesize():
                res.append((ofi.filename(),"resized",ofi.filesize(),"->",fi.filesize()))
                if options.timeline: self.timeline.append((fi.mtime(),fi.filename(),"resized",ofi.filesize(),"->",fi.filesize()))
            if ofi.sha1() != fi.sha1():
                res.append((ofi.filename(),"SHA1 changed",ofi.sha1(),"->",fi.sha1()))
                if options.timeline: self.timeline.append((fi.mtime(),fi.filename(),"SHA1 changed",ofi.sha1(),"->",fi.sha1()))
            if ofi.atime() != fi.atime():
                res.append((ofi.filename(),"atime changed",ofi.atime(),"->",fi.atime()))
                if options.timeline: self.timeline.append((fi.atime(),fi.filename(),"atime changed",prtime(ofi.atime()),"->",prtime(fi.atime())))
            if ofi.mtime() != fi.mtime():
                res.append((ofi.filename(),"mtime changed to",ofi.mtime(),"->",fi.mtime()))
                if options.timeline: self.timeline.append((fi.mtime(),fi.filename(),"mtime changed",prtime(ofi.mtime()),"->",prtime(fi.mtime())))
            if ofi.ctime() != fi.ctime():
                res.append((ofi.filename(),"ctime changed to",ofi.ctime(),"->",fi.ctime()))
                if options.timeline: self.timeline.append((fi.ctime(),fi.filename(),"ctime changed",prtime(ofi.ctime()),"->",prtime(fi.ctime())))
            if ofi.crtime() != fi.crtime():
                res.append((ofi.filename(),"crtime changed to",ofi.crtime(),"->",fi.crtime()))
                if options.timeline: self.timeline.append((fi.crtime(),fi.filename(),"crtime changed",prtime(ofi.crtime()),"->",prtime(fi.crtime())))
        if res:
            res.sort()
            table(res)

    def print_timeline(self):
        prt = []
        self.timeline.sort()
        # Make the dates printable
        for line in self.timeline:
            prt.append([ptime(line[0])]+list(line[1:]))
        h2("Timeline")
        table(prt)

    def report(self):
        header()
        h1("Disk image:"+self.current_fname)
        self.print_fis("New Files:",self.new_files)
        self.print_fis("Deleted Files:",self.fnames.values())
        self.print_fi2("Files with modified content:",self.changed_content)
        self.print_fi2("Files with changed file properties:",self.changed_properties)
        if options.timeline: self.print_timeline()

    def outputtar(self,infile,tarname=None):
        import tarfile
        import StringIO

        found_one = False
        if not tarname:
            tarname = self.current_fname+".tar"
        print ">>> Creating tar file: %s" % tarname
        tar = tarfile.TarFile(tarname,"w")
        changed = self.new_files
        for (ofi, fi) in self.changed_content + self.changed_properties:
            changed.append(fi)

        for fo in changed:
           if not fo.allocated() : continue
           if fo.has_tag("inode") and fo.has_contents():
               found_one = True
               info = tarfile.TarInfo(name=fo.filename())
               info.mtime = fo.mtime()
               info.atime = fo.atime()
               info.ctime = fo.ctime()
               info.uid = fo.uid()
               info.gid = fo.gid()
               string = StringIO.StringIO()
               string.write(fo.contents(imagefile=open(infile)))
               string.seek(0)
               info.size = fo.filesize()
               tar.addfile(tarinfo=info,fileobj=string)
        tar.close()
        if not found_one: unlink(tarname)  #remove incomplete file

if __name__=="__main__":
    from optparse import OptionParser
    from copy import deepcopy
    global options

    parser = OptionParser()
    parser.usage = '%prog [options] [file1 file2 file3]  (files can be xml or image files)'
    parser.add_option("-x","--xml",help="specify output file for XML",dest="xmlfilename")
    parser.add_option("--html",help="specify output in HTML",action="store_true")
    parser.add_option("-n","--notimeline",help="do not generate a timeline",action="store_true")
    parser.add_option("-d","--debug",help="debug")
    parser.add_option("-T","--tararchive",help="create tar archive file from new/changed files",dest="tarfile")

    (options,args) = parser.parse_args()
    options.timeline = True

    if len(args)<1:
        parser.print_help()
        sys.exit(1)

    s = DiskState()
    first = True
    for infile in args:
        print ">>> Reading %s" % infile
        s.process(infile)
        if not first:
            if options.tarfile != None:
                s.outputtar(infile,options.tarfile)
            s.report()
        first = False
        s.next()




